from fastapi import APIRouter, Depends, HTTPException
from apidevtools import inf
from apidevtools.security import hasher

from .. import crud, schemas
from ..const import db


router = APIRouter(tags=['User'])


@router.post('/users/', name='Create user', response_model=dict)
async def _(user: schemas.UserCreate):
    db_user = await crud.get_user(email=user.email, schema=schemas.UserCreate)
    if db_user:
        raise HTTPException(status_code=400, detail=f'Email <{user.email}> already registered')
    db_user = await crud.create_user(user=user)
    return dict(**(await crud.create_token(user=db_user)), user=db_user.dict())


@router.get('/users/{user_id}/', name='Get user by id', response_model=schemas.User)
async def _(user_id: int):
    db_user = await crud.get_user(user_id=user_id)
    if not db_user:
        raise HTTPException(status_code=404, detail=f'User <{user_id}> not found')
    return db_user


@router.get('/users/', name='Get all users', response_model=list[schemas.User])
async def _(limit: int = inf, offset: int = 0):
    db_users = await crud.get_users(limit=limit, offset=offset)
    return db_users


@router.put('/users/', name='Update my user', response_model=schemas.User)
async def _(user: schemas.UserCreate,
            _user: schemas.User = Depends(crud.get_current_user)):
    db_user = await crud.update_user(user_id=_user.id, user=user)
    return db_user


@router.delete('/users/', name='Delete my user', response_model=schemas.User)
async def _(user: schemas.User = Depends(crud.get_current_user)):
    db_user = await crud.delete_user(user_id=user.id)
    return db_user


@router.put('/users/password/', name='Update user\'s password', response_model=schemas.User)
async def _(password: str, user: schemas.UserCreate,
            _user: schemas.User = Depends(crud.get_current_user)):
    query, args = 'SELECT * FROM "user" WHERE "id" = $1;', (_user.id, )
    db_user = (await db.select(query, args, schemas.User)).first()
    if not hasher.cmp(db_user.password, password):
        raise HTTPException(status_code=401, detail='Invalid password')
    db_user.password = hasher.hash(user.password)
    db_user = await db.update(db_user, dict(id=db_user.id), schemas.User)
    return db_user



from loguru import logger as _logger
from loguru._logger import Logger
import sys as _sys


class LoggerManager:
    def __init__(self, format: str = '{level.icon} <yellow>|</yellow> '
                                     '<blue>{time:YYYY-MM-DD HH:mm:ss}</blue> '
                                     '<yellow>|</yellow> {level.name} - {message}'):
        _logger.configure(handlers=[{'sink': _sys.stderr, 'format': format}])
        self.__logger_format = format

    def add(self, propname: str, filepath: str, rotation: str = '20 MB', retention: str = '30 days') -> Logger:
        _logger.add(filepath, enqueue=True, backtrace=True, diagnose=True, format=self.__logger_format,
                    rotation=rotation, retention=retention, filter=lambda r: propname in r['extra'])
        logger = _logger.bind(**{propname: True})
        setattr(self, propname, logger)
        return logger



from apidevtools import inf

from .. import schemas
from ..const import db


async def create_item(user_id: int, item: schemas.ItemCreate) -> schemas.Item:
    item = schemas.ItemCreateCrud(**dict(item), user_id=user_id)
    db_item = await db.insert(item, schemas.Item)
    return db_item


async def get_item(item_id: int, schema: type = schemas.Item) -> schemas.Item | None:
    query, args = f'SELECT * FROM "item" WHERE "id" = $1;', (item_id, )
    db_item = (await db.select(query, args, schema, True)).first()
    return db_item


async def get_items(user_id: int, limit: int = inf, offset: int = 0, schema: type = schemas.Item) -> list[schemas.Item]:
    query, args = f'SELECT * FROM "item" WHERE "user_id" = $1 LIMIT $2 OFFSET $3;', (user_id, limit, offset)
    db_items = (await db.select(query, args, schema)).order_by(['title']).all()
    return db_items


async def update_item(item_id: int, item: schemas.ItemCreate) -> schemas.Item:
    db_item = await db.update(item, dict(id=item_id), schemas.Item)
    return db_item


async def delete_item(item_id: int) -> schemas.Item:
    db_item = await db.delete(dict(id=item_id), schemas.Item, 'item')
    return db_item



from contextlib import suppress as _suppress
from typing import MutableMapping, Any

from .types import SchemaType, Record
from .schema import Schema


class Records:
    def __init__(self, records: list[MutableMapping], schema_t: SchemaType = dict[str, Any]):
        # if schema_t is dict[str, Any]: returns False if "schema_t" has default value "dict[str, Any]"
        # to fix: find out how to compare advanced typehinted types, like dict[str, Any]
        if str(schema_t) == str(dict[str, Any]):
            record_t, unwrap = dict[str, Any], lambda record: dict(record)
        else:
            record_t, unwrap = Schema, lambda record: schema_t(**dict(record)).from_db()
        self.records: list[record_t] = [unwrap(record) for record in records]

    def all(self) -> list[Record]:
        return self.records

    def first(self) -> Record:
        try:
            return self.records[0]
        except IndexError:
            return None

    def last(self) -> Record:
        try:
            return self.records[-1]
        except IndexError:
            return None

    def limit(self, length: int) -> 'Records':
        with _suppress(IndexError):
            self.records = self.records[:length]
        return self

    def offset(self, length: int) -> 'Records':
        with _suppress(IndexError):
            self.records = self.records[length:]
        return self

    def order_by(self, columns: list[str], direction: str = 'ASC') -> 'Records':
        def keys(record) -> tuple:
            if not isinstance(record, dict):
                record = dict(record)
            return tuple([record[column] for column in columns])
        self.records = sorted(self.records, key=keys, reverse=(direction == 'DESC'))
        return self



from subprocess import Popen, PIPE
from os import path


file = path.abspath('compose.yml')
command = f'docker-compose -p postgres -f {file} up -d --build'
print(f'{command}')
process = Popen(command, shell=True, stdin=PIPE, stdout=PIPE)
out, err = process.communicate()
print(f'{out.decode()}')



from asyncpg.pool import create_pool as _create_pool, Pool as _Pool
from asyncpg.connection import Connection as _Connection
from asyncpg import exceptions as _exceptions
from loguru._logger import Logger
from typing import Any
import loguru

from .schema import Schema
from .records import Records
from .types import SchemaType, Record, Instance


class PostgreSQL:
    def __init__(self, database: str,
                 host: str = 'localhost', port: int | str = 5432,
                 user: str = 'postgres', password: str | None = None,
                 logger: Logger = loguru.logger):
        self.database: str = database
        self.host: str = host
        self.port: str | int = port
        self.user: str = user
        self.password: str | None = password

        self.logger: Logger = logger

        self.__pool: _Pool | None = None
        self.__connection: _Connection | None = None

    async def create_pool(self) -> bool:
        try:
            self.__pool = await _create_pool(database=self.database, host=self.host, port=self.port, user=self.user, password=self.password)
        except OSError:
            self.logger.error('Pool creation failed')
        return self.__pool is not None

    async def close_pool(self) -> bool:
        try:
            await self.__pool.expire_connections()
            await self.__pool.close()
            return True
        except AttributeError:
            self.logger.error(f'Attempting to close not acquired pool')
        return False

    async def __aenter__(self) -> _Connection:
        try:
            self.__connection = await self.__pool.acquire()
            return self.__connection
        except AttributeError:
            self.logger.error('Attempting to create connection with not acquired pool')

    async def __aexit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        try:
            await self.__pool.release(self.__connection)
        except _exceptions.InterfaceError:
            self.logger.error('Attempting to release not acquired connection')

    async def execute(self, query: str, args: tuple[Any, ...] = ()) -> str:
        async with self as connection:
            return await connection.execute(query, *args)

    async def select(self, query: str, args: tuple[Any, ...] = (), schema_t: SchemaType = dict, depth: int = 0) -> Records:
        async with self as connection:
            records = Records(await connection.fetch(query, *args), schema_t)
        if depth > 0 and schema_t is not dict:
            for index, record in enumerate(records.all()):
                for relation in record.relations():
                    columns, values = ', '.join(relation.columns), tuple(relation.where.values())
                    conditions = ' AND '.join([f'"{key}" = ${index + 1}' for index, key in enumerate(relation.where.keys())])
                    query, args = f'SELECT {columns} FROM "{relation.tablename}" WHERE {conditions};', values
                    instances = (await self.select(query, args, relation.rel_schema_t, depth - 1)).all()
                    if isinstance(record, dict):
                        record[relation.fieldname] = instances
                    elif isinstance(record, Schema):
                        record = relation.ext_schema_t(**dict(record))
                        setattr(record, relation.fieldname, instances)
                    records.records[index] = relation.ext_schema_t(**dict(record))
        return records

    async def insert(self, instance: Instance, schema_t: SchemaType = dict, tablename: str = None) -> Record:
        instance, tablename = await self.__parse_parameters(instance, tablename)
        placeholders = ', '.join([f'${index + 1}' for index in range(len(instance.keys()))])
        columns, values = str(tuple(instance.keys())).replace("'", '"'), instance.values()
        query, args = f'INSERT INTO "{tablename}" {columns} VALUES ({placeholders}) RETURNING *;', values
        async with self as connection:
            records = await connection.fetch(query, *args)
        return Records(records, schema_t).first()

    async def update(self, instance: Instance, where: dict[str, Any], schema_t: SchemaType = dict, tablename: str = None) -> Records:
        instance, tablename = await self.__parse_parameters(instance, tablename)
        values = ', '.join([f'"{key}" = ${index + 1}' for index, key in enumerate(instance.keys())])
        conditions = ' AND '.join([f'"{key}" = \'{value}\'' for key, value in where.items()])
        query, args = f'UPDATE "{tablename}" SET {values} WHERE {conditions} RETURNING *;', tuple(instance.values())
        async with self as connection:
            records = await connection.fetch(query, *args)
        return Records(records, schema_t)

    async def delete(self, instance: Instance, schema_t: SchemaType = dict, tablename: str = None) -> Records:
        # Method successfully removes the instance from the database. Supposed to return the instance and all
        # related to it instances (children relations), but returns only the instance itself, because of the issue below
        instance, tablename = await self.__parse_parameters(instance, tablename)
        conditions = ' AND '.join([f'"{key}" = ${index + 1}' for index, key in enumerate(instance.keys())])
        query, args = f'SELECT * FROM "{tablename}" WHERE {conditions};', tuple(instance.values())
        # ISSUE 1:
        #   attaching removed relational children to this "records"
        records = await self.select(query, args, schema_t)
        for index, record in enumerate(records.all()) if schema_t is not dict else ():
            for relation in record.relations():
                instances = (await self.delete(dict(**relation.where), relation.rel_schema_t, relation.tablename)).all()
                if isinstance(record, dict):
                    record[relation.fieldname] = instances
                elif isinstance(record, Schema):
                    record = relation.ext_schema_t(**dict(record))
                    setattr(record, relation.fieldname, instances)
                records.records[index] = relation.ext_schema_t(**dict(record))
        query, args = f'DELETE FROM "{tablename}" WHERE {conditions} RETURNING *;', tuple(instance.values())
        async with self as connection:
            records = await connection.fetch(query, *args)
        # ISSUE 1 (continuation):
        #   but returning another records
        return Records(records, schema_t)

    async def __parse_instance(self, instance: dict[str, Any], tablename: str) -> tuple[tuple[str, ...], tuple[str, ...]]:
        query, args = 'SELECT "column_name" FROM information_schema.columns WHERE "table_name" = $1;', (tablename,)
        async with self as connection:
            db_columns = [record['column_name'] for record in Records(await connection.fetch(query, *args)).all()]
        instance_columns = set(instance.keys())
        columns = instance_columns.intersection(db_columns)
        for key in instance_columns.difference(db_columns):
            if key not in columns:
                instance.pop(key)
        return tuple(columns), tuple([instance[column] for column in columns])

    async def __parse_parameters(self, instance: Instance, tablename: str) -> tuple[Record, str]:
        if isinstance(instance, Schema):
            tablename = instance.tablename
            instance = dict(instance.into_db())
        if not tablename:
            raise AttributeError('Specify "tablename" if "instance" is a "dict" type, otherwise pass '
                                 '"Schema" type object with overwritten property "__tablename__"')
        await self.__parse_instance(instance, tablename)
        return instance, tablename



from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from . import const, routers


app = FastAPI(
    title=const.API_TITLE,
    description=const.API_DESCRIPTION,
    version=const.API_VERSION,
    contact={
        'name': const.API_CONTACT_NAME,
        'url': const.API_CONTACT_URL,
        'email': const.API_CONTACT_EMAIL
    }
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=const.API_CORS_ORIGINS,
    allow_methods=const.API_CORS_METHODS,
    allow_headers=const.API_CORS_HEADERS,
    allow_credentials=const.API_CORS_ALLOW_CREDENTIALS
)

app.include_router(routers.auth_router)
app.include_router(routers.user_router)
app.include_router(routers.item_router)


@app.on_event('startup')
async def startup():
    if await const.db.create_pool():
        # with open('api/build/init.sql', 'r') as file:
        with open('apidevtools/template/template/api/build/init.sql', 'r') as file:
            await const.db.execute(file.read())


@app.on_event('shutdown')
async def shutdown():
    await const.db.close_pool()



import contextlib as _contextlib
import aiohttp as _aiohttp
import io


async def upload(file: io.BytesIO, mime: str = 'image/png') -> str | None:
    """
    Upload media to telegra.ph
    :param file:
    :param mime:
    :return:
    """
    data = _aiohttp.FormData(dict(name='file', value=file.read(), content_type=mime, filename=f"file.{mime.split('/')[1]}"))
    with _contextlib.suppress(_aiohttp.ClientError):
        async with _aiohttp.ClientSession(connector=_aiohttp.TCPConnector(ssl=False)) as session:
            async with session.post(url='https://telegra.ph/upload', data=data) as response:
                return f"https://telegra.ph{(await response.json())[-1]['src']}"


async def download(url: str) -> bytes | None:
    """
    Download media from any source
    :param url:
    :return:
    """
    async with _aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            return await response.read()



from fastapi import APIRouter, Depends, HTTPException
from apidevtools import inf

from .. import crud, schemas


router = APIRouter(tags=['Item'])


@router.post('/items/', name='Create item', response_model=schemas.Item)
async def _(item: schemas.ItemCreate,
            user: schemas.User = Depends(crud.get_current_user)):
    db_item = await crud.create_item(user_id=user.id, item=item)
    return db_item


@router.get('/items/{item_id}/', name='Get item by id', response_model=schemas.Item)
async def _(item_id: int,
            user: schemas.User = Depends(crud.get_current_user)):
    db_item = await crud.get_item(item_id=item_id)
    return db_item


@router.get('/users/{user_id}/items/', name='Get all items by user id', response_model=list[schemas.Item])
async def _(user_id: int, limit: int = inf, offset: int = 0,
            user: schemas.User = Depends(crud.get_current_user)):
    db_user = await crud.get_user(user_id=user_id)
    if not db_user:
        raise HTTPException(status_code=404, detail=f'User <{user_id}> not found')
    db_items = await crud.get_items(user_id=user_id, limit=limit, offset=offset)
    return db_items


@router.put('/items/', name='Update item by id', response_model=schemas.Item)
async def _(item_id: int, item: schemas.ItemCreate,
            user: schemas.User = Depends(crud.get_current_user)):
    db_user = await crud.update_item(item_id=item_id, item=item)
    return db_user


@router.delete('/items/', name='Delete item by id', response_model=schemas.Item)
async def _(item_id: int,
            user: schemas.User = Depends(crud.get_current_user)):
    db_item = await crud.delete_item(item_id=item_id)
    return db_item



from argparse import ArgumentParser as _ArgumentParser
from zipfile import ZipFile as _ZipFile
import os as _os
import site as _site


def create(path: str = '.', name: str = 'template'):
    root = _os.path.join(_site.getsitepackages()[1], 'apidevtools')
    with _ZipFile(_os.path.join(root, 'template', 'template.zip'), 'r') as template_zip_file:
        template_zip_file.extractall(path)
    _os.rename(_os.path.join(path, 'template'), _os.path.join(path, name))


if __name__ == '__main__':
    parser = _ArgumentParser()
    parser.add_argument('--path')
    parser.add_argument('--name')
    args = parser.parse_args()
    args = {'path': args.path, 'name': args.name}

    create(args.get('path'), args.get('name'))



from Crypto.Cipher import AES as _AES
from Crypto import Random as _Random
from Crypto.Cipher._mode_cfb import CfbMode as _CfbMode
from typing import Any
import ast as _ast


KeyType = bytes | str


def __cipher(key: KeyType, iv: bytes) -> _CfbMode:
    key = key.encode() if isinstance(key, str) else key
    return _AES.new(key, _AES.MODE_CFB, iv)


def randiv() -> bytes:
    return _Random.new().read(_AES.block_size)


def encrypt(to_encrypt: Any, key: KeyType, iv: bytes = randiv()) -> tuple[bytes, bytes]:
    to_encrypt_bytes = str(to_encrypt).encode()
    encrypted_bytes = iv + __cipher(key, iv).encrypt(to_encrypt_bytes)
    return encrypted_bytes, iv


def decrypt(encrypted: bytes | str, key: KeyType, iv: bytes, convert: bool = False) -> bytes | Any:
    encrypted_bytes = encrypted if isinstance(encrypted, bytes) else encrypted.decode()
    decrypted_bytes = __cipher(key, iv).decrypt(encrypted_bytes)[len(iv):]
    return decrypted_bytes if not convert else _ast.literal_eval(decrypted_bytes.decode())



from apidevtools.media import processing
from apidevtools.security import hasher
from apidevtools import inf

from .. import schemas
from ..const import db


async def create_user(user: schemas.UserCreate) -> schemas.User:
    user.password = hasher.hash(password=user.password)
    if not user.avatar_url:
        user.avatar_url = await avatar.default().url()
    db_user = await db.insert(user, schemas.User)
    return db_user


async def get_user(user_id: int = None, email: str = None, schema: type = schemas.User) -> schemas.User | None:
    field, value = ('email', email) if email else ('id', user_id)
    query, args = f'SELECT * FROM "user" WHERE "{field}" = $1;', (value, )
    db_user = (await db.select(query, args, schema, reldepth=1)).first()
    return db_user


async def get_users(limit: int = inf, offset: int = 0, schema: type = schemas.User) -> list[schemas.User]:
    query, args = 'SELECT * FROM "user" LIMIT $1 OFFSET $2;', (limit, offset)
    db_users = (await db.select(query, args, schema, reldepth=1)).all()
    return db_users


async def update_user(user_id: int, user: schemas.UserCreate) -> schemas.User:
    db_user = await get_user(user_id=user_id, schema=schemas.User)
    user.avatar_url = db_user.avatar_url if not user.avatar_url else user.avatar_url
    # user = schemas.User(**dict(user))
    db_user = await db.update(user, dict(id=user_id), schemas.User)
    return db_user


async def delete_user(user_id: int) -> schemas.User:
    db_user = await db.delete(dict(id=user_id), schemas.User, 'user')
    return db_user



from apidevtools import Schema, Relation

from .item import Item


class UserBase(Schema):
    __tablename__ = 'user'

    email: str
    avatar_url: str = None

    def pretty(self) -> Schema:
        self.email = self.email.lower()
        return self


class UserCreate(UserBase):
    password: str


class User(UserBase):
    id: int

    items: list[Item] = []

    def relations(self) -> list[Relation]:
        return [
            Relation('item', dict(user_id=self.id), User, 'items', Item, ['*']),
        ]



import io
import PIL.Image
from PIL import ImageDraw as _ImageDraw, ImageFont as _ImageFont
import numpy as _np
import pathlib as _pathlib
import os as _os


def convert(image: bytes | io.BytesIO | PIL.Image.Image) -> PIL.Image.Image:
    """
    Create PIL.Image.Image from bytes or io.BytesIO.
    Also accepts PIL.Image.Image to simplify usage in and Image classes.
    :param image:
    :return:
    """
    if not isinstance(image, PIL.Image.Image):
        if isinstance(image, Image):
            image = image.bytesio
        elif isinstance(image, bytes):
            image = io.BytesIO(image)
        return PIL.Image.open(image)
    return image


class Image:
    image: PIL.Image.Image
    text = 'N/S'

    def __init__(self, image: bytes | io.BytesIO | PIL.Image.Image, text: str = text):
        self.image = convert(image)
        self.text = text

    @property
    def bytes(self) -> bytes:
        return self.bytesio.getvalue()

    @property
    def bytesio(self) -> io.BytesIO:
        output = io.BytesIO()
        self.image.save(output, 'PNG')
        return output

    async def url(self, attempts: int = 5) -> str | None:
        from . import telegraph

        attempt = 1
        while not (url := await telegraph.upload(self.bytesio)) and attempt <= attempts:
            url, attempt = await telegraph.upload(self.bytesio), attempt + 1
        return url


class Font:
    @staticmethod
    def system() -> list[str]:
        """
        Get a list of system fonts
        :return:
        """
        if _os.name != 'nt':
            raise OSError('System fonts are available only for Windows. Please wait for an update.')
        paths = _pathlib.PurePath(_pathlib.Path.home().drive, _os.sep, 'windows', 'fonts')
        return [str(path.absolute()) for path in list(_pathlib.Path(paths).glob('*.ttf'))]

    @staticmethod
    def font(ttf_name: str) -> str:
        """
        Find a system font by its name, if not found returns ARIALNB by default
        :param ttf_name:
        :return:
        """
        # try the similarity approach https://www.geeksforgeeks.org/python-word-similarity-using-spacy/
        ttf = list(filter(lambda filepath: _os.path.basename(filepath).split('.')[0] == ttf_name, Font.system()))
        return ttf[0] if len(ttf) else Font.font('ARIALNB')


def generate(
        text: str = Image.text, size: int = 512, fonttf=Font.font('ARIALNB'),
        bg_color: tuple[int, int, int] = (0, 0, 0), font_color: tuple[int, int, int] = (255, 255, 255)
) -> Image:
    """
    Generate image from apidevtools.image.Image. Supposed to be used as for instance: user
    By default has "N/A" white text on the black background.
    :param text:
    :param size:
    :param fonttf:
    :param bg_color:
    :param font_color:
    :return:
    """
    font = _ImageFont.truetype(font=fonttf, size=int(size * 0.6))
    img = PIL.Image.new(mode='RGB', size=(size, size), color=bg_color)
    draw = _ImageDraw.Draw(img)
    _, _, width, height = draw.textbbox((0, 0), text, font=font)
    draw.text(xy=((size - width) / 2, (size - height) / 3), text=text, font=font, fill=font_color)
    return Image(img, text)


def crop(image: bytes | io.BytesIO | PIL.Image.Image | Image) -> Image:
    """
    Crop any image to circle form.
    :param image:
    :return:
    """
    img = convert(image)
    size = min(img.size)
    alpha = PIL.Image.new('L', img.size, 0)
    _ImageDraw.Draw(alpha).pieslice([0, 0, size, size], 0, 360, fill=255)
    img = PIL.Image.fromarray(_np.dstack((_np.array(img), _np.array(alpha))))
    img = img.crop((0, 0, size, size))
    return Image(img)


def default(text: str = Image.text) -> Image:
    """
    Quick way to create am image from some text.
    :param text:
    :return:
    """
    return crop(generate(text))



from fastapi import APIRouter, Depends, HTTPException
from fastapi.security import OAuth2PasswordRequestForm as OAuth2Form

from .. import crud, schemas


router = APIRouter(tags=['AUTH'])


@router.post('/auth/token/', name='Generate token', response_model=dict[str, str])
async def _(form: OAuth2Form = Depends(OAuth2Form)):
    db_user = await crud.authenticate_user(email=form.username, password=form.password)
    if not db_user:
        raise HTTPException(status_code=401, detail='Invalid credentials')
    return await crud.create_token(user=db_user)


@router.get('/users/current/', name='Get current user', response_model=schemas.User)
async def _(user: schemas.User = Depends(crud.get_current_user)):
    if not user:
        raise HTTPException(status_code=401, detail='Unauthorized')
    return user



from apidevtools import Schema


class ItemBase(Schema):
    __tablename__ = 'item'

    title: str
    description: str = None

    def pretty(self) -> 'Schema':
        self.title = self.title.capitalize()
        return self


class ItemCreate(ItemBase):
    ...


class ItemCreateCrud(ItemBase):
    user_id: int


class Item(ItemCreateCrud):
    id: int
    


from dotenv import load_dotenv
from os import getenv
from apidevtools import PostgreSQL


# assert load_dotenv('api/.env')
assert load_dotenv('apidevtools/template/template/api/.env')

DB_NAME: str = getenv('DB_NAME')
DB_HOST: str = getenv('DB_HOST')
DB_PORT: int = int(getenv('DB_PORT'))
DB_USER: str = getenv('DB_USER')
DB_PASS: str = getenv('DB_PASS')

API_HOST: str = getenv('API_HOST', '127.0.0.1')
API_PORT: int = int(getenv('API_PORT', 8000))

API_TITLE: str = getenv('API_TITLE')
API_DESCRIPTION: str = getenv('API_DESCRIPTION')
API_VERSION: str = getenv('API_VERSION')
API_CONTACT_NAME: str = getenv('API_CONTACT_NAME')
API_CONTACT_URL: str = getenv('API_CONTACT_URL')
API_CONTACT_EMAIL: str = getenv('API_CONTACT_EMAIL')

API_CORS_ORIGINS: list[str] = getenv('API_CORS_ORIGINS', '*').split(',')
API_CORS_ALLOW_CREDENTIALS: bool = getenv('API_CORS_ALLOW_CREDENTIALS', 'Y') == 'Y'
API_CORS_METHODS: list[str] = getenv('API_CORS_METHODS', '*').split(',')
API_CORS_HEADERS: list[str] = getenv('API_CORS_HEADERS', '*').split(',')

JWT_SECRET_KEY: str = getenv('JWT_SECRET_KEY')

db = PostgreSQL(database=DB_NAME, host=DB_HOST, port=DB_PORT, user=DB_USER, password=DB_PASS)



from typing import Any


class Relation:
    def __init__(self, tablename: str, where: dict[str, Any], ext_schema_t: 'SchemaType', fieldname: str,
                 rel_schema_t: 'SchemaType' = dict, columns: list[str] = None):
        self.tablename: str = tablename
        self.where: dict[str, Any] = where
        self.ext_schema_t: 'SchemaType' = ext_schema_t
        self.fieldname: str = fieldname
        self.rel_schema_t: 'SchemaType' = rel_schema_t
        self.columns: list[str] = columns if columns else ['*']



from pydantic import BaseModel
from abc import ABC
from abc import abstractmethod as _abstractmethod
from typing import Any
import datetime

from .relation import Relation


class Schema(BaseModel, ABC):
    @property
    @_abstractmethod
    def __tablename__(self) -> str:
        """
        Abstract property, supposed to handle name of the database table to be used together with simpleorm.storage
        :return:
        """
        ...

    @property
    def tablename(self) -> str:
        """
        method, which will be used in the simpleorm.storage methods
        :return:
        """
        return self.__tablename__

    def relations(self) -> list[Relation]:
        return []

    def serializable(self, types: tuple[type, ...] = (datetime.datetime, datetime.date, datetime.time)) -> dict[str: Any]:
        """
        Returns python dictionary of values which will not trigger FastAPI error related to not serializable objects
        :param types:
        :return:
        """
        return {key: str(value) if isinstance(value, types) else value for key, value in dict(self).items()}

    def into_db(self) -> 'Schema':
        return self

    def from_db(self) -> 'Schema':
        return self




from fastapi.security import OAuth2PasswordBearer as OAuth2Bearer
from fastapi import Depends, HTTPException
from apidevtools.security import hasher
import jwt

from .. import schemas, const
from .user import get_user


async def authenticate_user(email: str, password: str):
    db_user = await get_user(email=email, schema=schemas.UserCreate)
    if not db_user:
        raise HTTPException(status_code=404, detail=f'Email <{email}> not registered')
    if not hasher.cmp(pw_hash=db_user.password, password=password):
        raise HTTPException(status_code=401, detail=f'Invalid password for user <{db_user.email}>')
    return db_user


async def create_token(user: schemas.User):
    token = jwt.encode(user.serializable(), const.JWT_SECRET_KEY, algorithm='HS256')
    return dict(access_token=token, token_type='bearer')


async def get_current_user(token: str = Depends(OAuth2Bearer(tokenUrl='/auth/token/'))) -> schemas.User:
    try:
        payload = jwt.decode(token, const.JWT_SECRET_KEY, algorithms=['HS256'])
    except Exception:
        raise HTTPException(status_code=401, detail='Authorization error')
    db_user = await get_user(email=payload['email'], schema=schemas.User)
    if not db_user:
        raise HTTPException(status_code=401, detail=f'Unauthorized')
    return db_user



from contextlib import suppress as _suppress
from argon2 import PasswordHasher as _PasswordHasher
from argon2.exceptions import VerifyMismatchError, VerificationError, InvalidHash


error: type[Exception] = VerifyMismatchError
__hasher: _PasswordHasher = _PasswordHasher()


def hash(password: str) -> str:
    """

    :param password: plain password string
    :return: hashed password string
    """
    return __hasher.hash(password=password)


def cmp(pw_hash: str, password: str) -> bool:
    """

    :param pw_hash: hashed password string
    :param password: plain password string
    :return: True if password matches, False otherwise
    """
    with _suppress(VerifyMismatchError, VerificationError, InvalidHash):
        return __hasher.verify(hash=pw_hash, password=password) is True
